//+------------------------------------------------------------------+
//|       WORKING VERSION OF A SCALABLE MULTI-LAYER-PERCEPTRON CLASS.|
//+------------------------------------------------------------------+
//|                                                    stephen NJUKI |
//|                                                     AUGUST, 2024 |
//+------------------------------------------------------------------+
#property version   "1.00"
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
#include <My\Cnorm.mqh> //Batch Normalization Class
#define __EPSILON 1e-8  // non-zero " learning-rate target"
//+------------------------------------------------------------------+
//| Adaptive Learning Rate Type Enumerator                           |
//+------------------------------------------------------------------+
enum Eadaptive
{  ADAPTIVE_NONE = -1,
   ADAPTIVE_GRADIENT = 0,
   ADAPTIVE_MOMENTUM = 1,
   ADAPTIVE_DELTA = 2,
};
//+------------------------------------------------------------------+
//| Multi-Layer-Perceptron Class Interface                           |
//+------------------------------------------------------------------+
class Cmlp
{
protected:
   matrix            weights[];
   vector            biases[];

   matrix            gradients[];
   vector            deltas[];

   matrix            gradients_1st_moment[];
   matrix            gradients_2nd_moment[];

   vector            deltas_1st_moment[];
   vector            deltas_2nd_moment[];

   matrix            sum_gradients[];
   matrix            sum_gradients_update[];

   vector            sum_deltas[];
   vector            sum_deltas_update[];

   vector            target;

   vector            hidden_outputs[];

   int               hidden_layers;

   ENUM_ACTIVATION_FUNCTION   activation;
   ENUM_LOSS_FUNCTION         loss;

   Enormalize        normalization;

   double            minimum_learning_rate;

   vector            normalization_beta[];
   vector            normalization_gamma[];

   vector            Hadamard(vector &A, vector &B);
   matrix            TransposeRow(vector &V);
   matrix            TransposeCol(vector &V);

   void              AdaptiveGradientUpdate(double LearningRate);

   void              AdaptiveMomentumUpdate(double LearningRate, double FirstDecayRate, double SecondDecayRate, int EpochIndex);
   void              AdaptiveDeltaUpdate(double DecayRate);

   void              MomentumUpdate(double &Parameter, double &Change, double &FirstMoment, double &SecondMoment, double LearningRate, double FirstDecayRate, double SecondDecayRate, int EpochIndex);
   void              DeltaUpdate(double &Parameter, double &Change, double &SumChange, double &SumChangeUpdate, double DecayRate);

public:
   vector            inputs;
   vector            output;
   bool              validated;
   Cnormalize        NORM;

   void              Get(vector &Target)
   {                 target.Init(0);
      target.Copy(Target);
   }

   double            Get(ENUM_REGRESSION_METRIC R)
   {                 return(target.RegressionMetric(output, R));
   }

   void              Set(vector &Inputs)
   {                 inputs.Init(0);
      inputs.Copy(Inputs);
   }

   bool              Get(string File, double &Criteria, datetime &Version, bool Log = false);
   bool              Set(string File, double Criteria, bool Log = false);

   void              Forward();
   void              Backward(double LearningRate = 0.1, Eadaptive AdaptiveType = ADAPTIVE_NONE, double DecayRate_A = 0.9, double DecayRate_B = 0.9, int EpochIndex = 1);

   void              Cmlp(int &Settings[], ENUM_ACTIVATION_FUNCTION Activation, ENUM_LOSS_FUNCTION Loss, vector &NormalizationBeta[], vector &NormalizationGamma[], Enormalize Normalization = NORMALIZE_NONE, double InitialWeight = 0.1, double InitialBias = 0.0, double MinimumLearningRate = __EPSILON)
   {  validated = false;
      int _settings =    ArraySize(Settings);
      if(_settings >= 2 && _settings <= USHORT_MAX && Settings[ArrayMinimum(Settings)] > 0 && Settings[ArrayMaximum(Settings)] < USHORT_MAX && MinimumLearningRate > 0.0 && MinimumLearningRate < 1.0)
      {  ArrayResize(weights, _settings - 1);
         ArrayResize(biases, _settings - 1);
         ArrayResize(gradients, _settings - 1);
         ArrayResize(deltas, _settings - 1);
         //
         ArrayResize(gradients_1st_moment, _settings - 1);
         ArrayResize(gradients_2nd_moment, _settings - 1);
         ArrayResize(deltas_1st_moment, _settings - 1);
         ArrayResize(deltas_2nd_moment, _settings - 1);
         //
         ArrayResize(sum_gradients, _settings - 1);
         ArrayResize(sum_gradients_update, _settings - 1);
         ArrayResize(sum_deltas, _settings - 1);
         ArrayResize(sum_deltas_update, _settings - 1);
         //
         ArrayResize(hidden_outputs, _settings - 2);
         //
         ArrayResize(normalization_beta, _settings - 1);
         ArrayResize(normalization_gamma, _settings - 1);
         //
         hidden_layers = _settings - 2;
         bool _norm_validated = true;
         for(int i = 0; i < _settings - 1; i++)
         {  weights[i].Init(Settings[i + 1], Settings[i]);
            weights[i].Fill(InitialWeight);
            biases[i].Init(Settings[i + 1]);
            biases[i].Fill(InitialBias);
            //
            gradients[i].Init(Settings[i + 1], Settings[i]);
            gradients[i].Fill(0.0);
            deltas[i].Init(Settings[i + 1]);
            deltas[i].Fill(0.0);
            //
            gradients_1st_moment[i].Init(Settings[i + 1], Settings[i]);
            gradients_1st_moment[i].Fill(0.0);
            gradients_2nd_moment[i].Init(Settings[i + 1], Settings[i]);
            gradients_2nd_moment[i].Fill(0.0);
            //
            deltas_1st_moment[i].Init(Settings[i + 1]);
            deltas_1st_moment[i].Fill(0.0);
            deltas_2nd_moment[i].Init(Settings[i + 1]);
            deltas_2nd_moment[i].Fill(0.0);
            //
            sum_gradients[i].Init(Settings[i + 1], Settings[i]);
            sum_gradients[i].Fill(0.0);
            sum_gradients_update[i].Init(Settings[i + 1], Settings[i]);
            sum_gradients_update[i].Fill(0.0);
            //
            sum_deltas[i].Init(Settings[i + 1]);
            sum_deltas[i].Fill(0.0);
            sum_deltas_update[i].Init(Settings[i + 1]);
            sum_deltas_update[i].Fill(0.0);
            //
            //If normalization is used there has to be a beta and gamma value
            //for each node on each layer, except the input layer.
            if(Normalization != NORMALIZE_NONE && (Settings[i + 1] != NormalizationBeta[i].Size() || Settings[i + 1] != NormalizationGamma[i].Size()))
            {  _norm_validated = false;
               break;
            }
            normalization_beta[i].Copy(NormalizationBeta[i]);
            normalization_gamma[i].Copy(NormalizationGamma[i]);
            //
            if(i < _settings - 2)
            {  hidden_outputs[i].Init(Settings[i + 1]);
               hidden_outputs[i].Fill(0.0);
            }
         }
         output.Init(Settings[_settings - 1]);
         target.Init(Settings[_settings - 1]);
         activation = Activation;
         loss = Loss;
         normalization = Normalization;
         minimum_learning_rate = MinimumLearningRate;
         if(_norm_validated)
         {  validated = true;
         }
      }
      else
      {  printf(__FUNCSIG__ +
                " invalid network settings! Settings size is: %i, Max layer size is: %i, Min layer size is: %i, and activation is %s ",
                _settings, Settings[ArrayMaximum(Settings)], Settings[ArrayMinimum(Settings)], EnumToString(Activation)
               );
      }
   };
   void              ~Cmlp(void) { };
};
//+------------------------------------------------------------------+
//| Vector to Matrix-Row Tranposer                                   |
//+------------------------------------------------------------------+
matrix Cmlp::TransposeRow(vector &V)
{  matrix _m;
   _m.Init(1, V.Size());
   _m.Fill(0.0);
   for(int i = 0; i < int(V.Size()); i++)
   {  _m[0][i] = V[i];
   }
   return(_m);
}
//+------------------------------------------------------------------+
//| Vector to Matrix-Column Tranposer                                |
//+------------------------------------------------------------------+
matrix Cmlp::TransposeCol(vector &V)
{  matrix _m;
   _m.Init(V.Size(), 1);
   _m.Fill(0.0);
   for(int i = 0; i < int(V.Size()); i++)
   {  _m[i][0] = V[i];
   }
   return(_m);
}
//+------------------------------------------------------------------+
//| Hadamard Vector Product                                          |
//+------------------------------------------------------------------+
vector Cmlp::Hadamard(vector &A, vector &B)
{  vector _c;
   _c.Init(A.Size());
   _c.Fill(0.0);
   if(A.Size() == B.Size() && B.Size() > 0)
   {  for(int i = 0; i < int(A.Size()); i++)
      {  _c[i] = A[i] * B[i];
      }
   }
   return(_c);
}
//+------------------------------------------------------------------+
//| FORWARD PROPAGATION THROUGH THE MULTI-LAYER-PERCEPTRON.          |
//+------------------------------------------------------------------+
//|                                                                  |
//| -Extra Validation check of MLP settings is performed at run-time.|
//|  Chcecking of 'validation' parameter should ideally be performed |
//|  at class instance initialisation.                               |
//|                                                                  |
//| -Input data is normalized if normalization type was selected at  |
//|  class instance initialisation.                                  |
//+------------------------------------------------------------------+
void Cmlp::Forward()
{  if(!validated)
   {  printf(__FUNCSIG__ + " invalid network settings! ");
      return;
   }
//
   for(int h = 0; h <= hidden_layers; h++)
   {  vector _output;
      _output.Init(output.Size());
      if(h == 0)
      {  vector _first_out = weights[h].MatMul(inputs);
         if(normalization != NORMALIZE_NONE)
         {  NORM.Transform(normalization, _first_out, _output, normalization_beta[h], normalization_gamma[h]);
         }
      }
      else if(h > 0)
      {  vector _next_out = weights[h].MatMul(hidden_outputs[h - 1]);
         if(normalization != NORMALIZE_NONE)
         {  NORM.Transform(normalization, _next_out, _output, normalization_beta[h], normalization_gamma[h]);
         }
      }
      _output += biases[h];
      if(h < hidden_layers)
      {  _output.Activation(hidden_outputs[h], activation);
      }
      else if(h == hidden_layers)
      {  _output.Activation(output, activation);
      }
   }
}
//+------------------------------------------------------------------+
//| BACKWARD PROPAGATION OF THE MULTI-LAYER-PERCEPTRON.              |
//+------------------------------------------------------------------+
//|                                                                  |
//| -Extra Validation check of MLP settings is performed at run-time.|
//|  Chcecking of 'validation' parameter should ideally be performed |
//|  at class instance initialisation.                               |
//|                                                                  |
//| -Run-time Validation of learning rate, decay rates and epoch     |
//|  index is performed as these are optimisable inputs.             |
//+------------------------------------------------------------------+
void Cmlp::Backward(double LearningRate = 0.1, Eadaptive AdaptiveType = ADAPTIVE_NONE, double DecayRate_A = 0.9, double DecayRate_B = 0.9, int EpochIndex = 1)
{  if(!validated)
   {  printf(__FUNCSIG__ + " invalid network settings! ");
      return;
   }
   if(LearningRate <= 0.0 || LearningRate >= 1.0)
   {  printf(__FUNCSIG__ + " Learning Rate out of bounds! Should be between 0.0 to +1.0 range. ");
      return;
   }
   if(DecayRate_A <= 0.0 || DecayRate_A >= 1.0)
   {  printf(__FUNCSIG__ + " Decay Rate A out of bounds! Should be between 0.0 to +1.0 range. ");
      return;
   }
   if(DecayRate_B <= 0.0 || DecayRate_B >= 1.0)
   {  printf(__FUNCSIG__ + " Decay Rate B out of bounds! Should be between 0.0 to +1.0 range. ");
      return;
   }
   if(EpochIndex < 0)
   {  printf(__FUNCSIG__ + " Epoch Index Should start from 1. ");
      return;
   }
//COMPUTE DELTAS
   vector _last, _last_derivative;
   _last.Init(inputs.Size());
   if(hidden_layers == 0)
   {  _last = weights[hidden_layers].MatMul(inputs);
   }
   else if(hidden_layers > 0)
   {  _last = weights[hidden_layers].MatMul(hidden_outputs[hidden_layers - 1]);
   }
   _last.Derivative(_last_derivative, activation);
   vector _last_error = output.LossGradient(target, loss);
   deltas[hidden_layers] = Hadamard(_last_error, _last_derivative);
   if(hidden_layers > 0)
   {  for(int h = hidden_layers - 1; h >= 0;  h--)
      {  vector _next, _next_derivative;
         _next.Init(weights[h].Rows());
         if(h - 1 >= 0)
         {  _next = weights[h].MatMul(hidden_outputs[h - 1]);
         }
         else
         {  _next = weights[h].MatMul(inputs);
         }
         _next.Derivative(_next_derivative, activation);
         matrix _next_weights;
         _next_weights.Copy(weights[h + 1]);
         _next_weights.Transpose();
         vector _next_product = deltas[h + 1].MatMul(_next_weights);
         deltas[h] = Hadamard(_next_product, _next_derivative);
      }
   }
//COMPUTE GRADIENTS
   for(int h = hidden_layers; h >= 0;  h--)
   {  if(h > 0)
      {  gradients[h] = TransposeCol(deltas[h]).MatMul(TransposeRow(hidden_outputs[h - 1]));
      }
      else if(h == 0)
      {  gradients[h] = TransposeCol(deltas[h]).MatMul(TransposeRow(inputs));
      }
   }
// UPDATE WEIGHTS AND BIASES
// Non-Adaptive
   if(AdaptiveType == ADAPTIVE_NONE)
   {  for(int h = hidden_layers; h >= 0;  h--)
      {  weights[h] -= LearningRate * gradients[h];
         biases[h] -= LearningRate * deltas[h];
      }
   }
// Adaptive
   else if(AdaptiveType != ADAPTIVE_NONE)
   {  if(AdaptiveType == ADAPTIVE_GRADIENT)
      {  AdaptiveGradientUpdate(LearningRate);
      }
      else if(AdaptiveType == ADAPTIVE_MOMENTUM)
      {  AdaptiveMomentumUpdate(LearningRate, DecayRate_A, DecayRate_B, EpochIndex);
      }
      else if(AdaptiveType == ADAPTIVE_DELTA)
      {  AdaptiveDeltaUpdate(DecayRate_A);
      }
   }
}
//+------------------------------------------------------------------+
//| Adaptive Gradient Update function                                |
//+------------------------------------------------------------------+
void Cmlp::AdaptiveGradientUpdate(double LearningRate)
{  for(int h = hidden_layers; h >= 0;  h--)
   {  for (int i = 0; i < int(weights[h].Rows()); i++)
      {  for (int j = 0; j < int(weights[h].Cols()); j++)
         {  sum_gradients[h][i][j] += (gradients[h][i][j] * gradients[h][i][j]);
            weights[h][i][j] -= ((LearningRate / (MathSqrt(sum_gradients[h][i][j]) + minimum_learning_rate)) * gradients[h][i][j]);
         }
         biases[h][i] -= LearningRate * deltas[h][i];
      }
   }
}
//+------------------------------------------------------------------+
//| Adaptive Momentum Estimation function                            |
//+------------------------------------------------------------------+
void Cmlp::AdaptiveMomentumUpdate(double LearningRate, double FirstDecayRate, double SecondDecayRate, int EpochIndex)
{  for(int h = hidden_layers; h >= 0;  h--)
   {  for (int i = 0; i < int(weights[h].Rows()); i++)
      {  for (int j = 0; j < int(weights[h].Cols()); j++)
         {  MomentumUpdate(weights[h][i][j], gradients[h][i][j], gradients_1st_moment[h][i][j], gradients_2nd_moment[h][i][j], LearningRate, FirstDecayRate, SecondDecayRate, EpochIndex);
         }
         MomentumUpdate(biases[h][i], deltas[h][i], deltas_1st_moment[h][i], deltas_2nd_moment[h][i], LearningRate, FirstDecayRate, SecondDecayRate, EpochIndex);
      }
   }
}
//+------------------------------------------------------------------+
//| Adaptive Delta Update function                                   |
//+------------------------------------------------------------------+
void Cmlp::AdaptiveDeltaUpdate(double DecayRate)
{  for(int h = hidden_layers; h >= 0;  h--)
   {  for (int i = 0; i < int(weights[h].Rows()); i++)
      {  for (int j = 0; j < int(weights[h].Cols()); j++)
         {  DeltaUpdate(weights[h][i][j], gradients[h][i][j], sum_gradients[h][i][j], sum_gradients_update[h][i][j], DecayRate);
         }
         DeltaUpdate(biases[h][i], deltas[h][i], sum_deltas[h][i], sum_deltas_update[h][i], DecayRate);
      }
   }
}
//+------------------------------------------------------------------+
//| Momentum Update method
//+------------------------------------------------------------------+
void Cmlp::MomentumUpdate(double &Parameter, double &Change, double &FirstMoment, double &SecondMoment, double LearningRate, double FirstDecayRate, double SecondDecayRate, int EpochIndex)
{  FirstMoment = (FirstDecayRate * FirstMoment) + ((1.0 - FirstDecayRate) * Change);
   SecondMoment = (SecondDecayRate * SecondMoment) + ((1.0 - SecondDecayRate) * Change * Change);
   double _1st_gradient = FirstMoment / (1.0 - pow(FirstDecayRate, EpochIndex));
   double _2nd_gradient = SecondMoment / (1.0 - pow(SecondDecayRate, EpochIndex));
   Parameter -= LearningRate * ((_1st_gradient + minimum_learning_rate) / (sqrt(_2nd_gradient) + minimum_learning_rate)) * Change;
}
//+------------------------------------------------------------------+
//| Delta Update method
//+------------------------------------------------------------------+
void Cmlp::DeltaUpdate(double &Parameter, double &Change, double &SumChange, double &SumChangeUpdate, double DecayRate)
{  SumChange = (DecayRate * SumChange * SumChange) + ((1.0 - DecayRate) * Change * Change);
   double _update = ((sqrt(SumChange) + minimum_learning_rate) / (sqrt(SumChangeUpdate) + minimum_learning_rate)) * Change;
   Parameter -= _update;
   SumChangeUpdate = (DecayRate * SumChangeUpdate * SumChangeUpdate) + ((1.0 - DecayRate) * _update * _update);
}
//+------------------------------------------------------------------+
//| Get Weights and Biases in import                                 |
//+------------------------------------------------------------------+
bool Cmlp::Get(string File, double &Criteria, datetime &Version, bool Log = false)
{  ResetLastError();
   int _handle = FileOpen(MQLInfoString(MQL_PROGRAM_NAME) + "_" + File + ".bin", FILE_READ | FILE_BIN | FILE_COMMON);
   if(_handle != INVALID_HANDLE)
   {  Criteria = FileReadDouble(_handle);
      Version = FileReadInteger(_handle);
      if(Log)
      {  PrintFormat(__FUNCSIG__ + " for the file: %s, we have the criteria: %.2f, logged on: %s ",
                     File, Criteria, TimeToString(Version));
      }
      for(int i = 0; i < int(weights.Size()); i++)
      {  for(int j = 0; j < int(weights[i].Rows()); j++)
         {  for(int k = 0; k < int(weights[i].Cols()); k++)
            {  weights[i][j][k] = FileReadDouble(_handle);
               if(Log)
               {  PrintFormat(__FUNCSIG__ +
                              " weight is: %.2f, for layer %i, at row: %i, and col: %i",
                              weights[i][j][k], i, j, k);
               }
            }
         }
      }
      for(int i = 0; i < int(biases.Size()); i++)
      {  for(int j = 0; j < int(biases[i].Size()); j++)
         {  biases[i][j] = FileReadDouble(_handle);
            if(Log)
            {  PrintFormat(__FUNCSIG__ +
                           " %.5f is the bias for layer %i, at row: %i",
                           biases[i][j], i, j);
            }
         }
      }
      FileClose(_handle);
   }
   else
   {  printf(__FUNCSIG__ + " failed to create get handle for: " + File + ", err: " + IntegerToString(GetLastError()));
      FileClose(_handle);
      return(false);
   }
   FileClose(_handle);
   return(true);
}
//+------------------------------------------------------------------+
//| Set Weights and Biases for export                                |
//+------------------------------------------------------------------+
bool Cmlp::Set(string File, double Criteria, bool Log = false)
{  ResetLastError();
   int _handle = FileOpen(MQLInfoString(MQL_PROGRAM_NAME) + "_" + File + ".bin", FILE_WRITE | FILE_BIN | FILE_COMMON);
   if(_handle != INVALID_HANDLE)
   {  FileWriteDouble(_handle, Criteria);
      FileWriteInteger(_handle, int(TimeCurrent()));
      if(Log)
      {  PrintFormat(__FUNCSIG__ + " for the file: %s, we have the criteria: %.2f, logged on: %s ",
                     File, Criteria, TimeToString(TimeCurrent()));
      }
      for(int i = 0; i < int(weights.Size()); i++)
      {  for(int j = 0; j < int(weights[i].Rows()); j++)
         {  for(int k = 0; k < int(weights[i].Cols()); k++)
            {  FileWriteDouble(_handle, weights[i][j][k]);
               if(Log)
               {  PrintFormat(__FUNCSIG__ +
                              " weight is: %.2f, for layer %i, at row: %i, and col: %i",
                              weights[i][j][k], i, j, k);
               }
            }
         }
      }
      for(int i = 0; i < int(biases.Size()); i++)
      {  for(int j = 0; j < int(biases[i].Size()); j++)
         {  FileWriteDouble(_handle, biases[i][j]);
            if(Log)
            {  PrintFormat(__FUNCSIG__ +
                           " %.5f is the bias for layer %i, at row: %i",
                           biases[i][j], i, j);
            }
         }
      }
      FileClose(_handle);
   }
   else
   {  printf(__FUNCSIG__ + " failed to create set handle for: " + File + ", err: " + IntegerToString(GetLastError()));
      FileClose(_handle);
      return(false);
   }
   FileClose(_handle);
   return(true);
}
//+------------------------------------------------------------------+
